"""IHC component constants."""

CONF_AUTOSETUP = 'auto_setup'
CONF_BINARY_SENSOR = 'binary_sensor'
CONF_DIMMABLE = 'dimmable'
CONF_INFO = 'info'
CONF_INVERTING = 'inverting'
CONF_LIGHT = 'light'
CONF_NAME = 'name'
CONF_NODE = 'node'
CONF_NOTE = 'note'
CONF_POSITION = 'position'
CONF_SENSOR = 'sensor'
CONF_SWITCH = 'switch'
CONF_XPATH = 'xpath'

ATTR_IHC_ID = 'ihc_id'
ATTR_VALUE = 'value'

SERVICE_SET_RUNTIME_VALUE_BOOL = 'set_runtime_value_bool'
SERVICE_SET_RUNTIME_VALUE_FLOAT = 'set_runtime_value_float'
SERVICE_SET_RUNTIME_VALUE_INT = 'set_runtime_value_int'
