"""Constants for the IGD component."""
import logging


CONF_ENABLE_PORT_MAPPING = 'port_mapping'
CONF_ENABLE_SENSORS = 'sensors'
CONF_HASS = 'hass'
CONF_LOCAL_IP = 'local_ip'
CONF_PORTS = 'ports'
DOMAIN = 'upnp'
LOGGER = logging.getLogger('homeassistant.components.upnp')
SIGNAL_REMOVE_SENSOR = 'upnp_remove_sensor'
